import express, { Request, Response } from 'express';
import jsonABC from 'jsonabc';
import equal from 'deep-equal';
import neo4jController from './controllers/neo4jController';

/** SwaggerDoc Config For Each Endpoint is Displayed Below */
const router = express.Router();

router.post('/LoadBundle', (req: Request, res: Response) => {
  return neo4jController.loadBundle(req.body, res);
});

router.post('/BuildBundle/:_id', (req: Request, res: Response) => {
  if (!req.params._id) {
    return res.status(400).send({
      message: 'ID Not Found in Database'
    });
  }
  return neo4jController.buildBundle(req.params._id, req.body.filter, res);
});

router.delete('/delete', (req: Request, res: Response) => {
  return neo4jController.deleteAll(req, res);
});

router.post('/compareJSON', (req: Request, res: Response) => {
  const sortedJsons = req.body.jsons.map((json) => jsonABC.sortObj(json));
  return res.status(200).send({ isEqual: equal(...sortedJsons) });
});

export default router;

/**
   * @swagger
   *
   *   /api/LoadBundle:
   *    post:
   *      summary: Load FHIR Bundle into Neo4j
   *      requestBody:
   *        required: true
   *        content:
   *           application/json:
   *             schema:
   *               type: object
   *             example: {"resourceType":"Bundle","type":"transaction","entry":[{"fullUrl":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","resource":{"resourceType":"Patient","id":"6aff2910-82fc-44d6-84a6-c29e4b756b11","text":{"status":"generated"},"extension":[{"url":"http://hl7.org/fhir/StructureDefinition/patient-mothersMaidenName","valueString":"Nydia801 Bashirian201"},{"url":"http://hl7.org/fhir/StructureDefinition/patient-birthPlace","valueAddress":{"city":"Monroe","state":"North Carolina","country":"US"}},{"url":"http://synthetichealth.github.io/synthea/disability-adjusted-life-years","valueDecimal":0.499751031865479},{"url":"http://synthetichealth.github.io/synthea/quality-adjusted-life-years","valueDecimal":56.50024896813452}],"identifier":[{"system":"https://github.com/synthetichealth/synthea","value":"6aff2910-82fc-44d6-84a6-c29e4b756b11"},{"type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v2-0203","code":"MR","display":"Medical Record Number"}],"text":"Medical Record Number"},"system":"http://hospital.smarthealthit.org","value":"6aff2910-82fc-44d6-84a6-c29e4b756b11"},{"type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v2-0203","code":"SS","display":"Social Security Number"}],"text":"Social Security Number"},"system":"http://hl7.org/fhir/sid/us-ssn","value":"999-14-9538"},{"type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v2-0203","code":"DL","display":"Driver's License"}],"text":"Driver's License"},"system":"urn:oid:2.16.840.1.113883.4.3.25","value":"S99997789"},{"type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v2-0203","code":"PPN","display":"Passport Number"}],"text":"Passport Number"},"system":"http://standardhealthrecord.org/fhir/StructureDefinition/passportNumber","value":"X57601805X"}],"name":[{"use":"official","family":"Haag279","given":["Theola421"],"prefix":["Mrs."]},{"use":"maiden","family":"Lehner980","given":["Theola421"],"prefix":["Mrs."]}],"telecom":[{"system":"phone","value":"555-196-9091","use":"home"}],"gender":"female","birthDate":"1962-09-10","address":[{"extension":[{"url":"http://hl7.org/fhir/StructureDefinition/geolocation","extension":[{"url":"latitude","valueDecimal":35.84622999061526},{"url":"longitude","valueDecimal":-79.9323600342269}]}],"line":["350 Schimmel Heights"],"city":"Archdale","state":"North Carolina","postalCode":"27263","country":"US"}],"maritalStatus":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-MaritalStatus","code":"M","display":"M"}],"text":"M"},"multipleBirthBoolean":false,"communication":[{"language":{"coding":[{"system":"urn:ietf:bcp:47","code":"en-US","display":"English"}],"text":"English"}}]},"request":{"method":"POST","url":"Patient"}},{"fullUrl":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8","resource":{"resourceType":"Encounter","id":"6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"185347001","display":"Encounter for problem"}],"text":"Encounter for problem"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2001-08-31T13:26:41-04:00","end":"2001-08-31T13:41:41-04:00"},"individual":{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0","display":"Dr. Latricia419 Spencer878"}}],"period":{"start":"2001-08-31T13:26:41-04:00","end":"2001-08-31T13:41:41-04:00"},"reasonCode":[{"coding":[{"system":"http://snomed.info/sct","code":"239873007","display":"Osteoarthritis of knee"}]}],"serviceProvider":{"reference":"urn:uuid:2864a135-87f9-316b-a6b7-a272636846b6","display":"HIGH POINT REGIONAL HOSPITAL"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:a4e96085-e006-4f9a-8eb9-00bc41b0ab15","resource":{"resourceType":"Condition","id":"a4e96085-e006-4f9a-8eb9-00bc41b0ab15","clinicalStatus":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/condition-clinical","code":"active"}]},"verificationStatus":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/condition-ver-status","code":"confirmed"}]},"code":{"coding":[{"system":"http://snomed.info/sct","code":"239873007","display":"Osteoarthritis of knee"}],"text":"Osteoarthritis of knee"},"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"},"onsetDateTime":"2001-08-31T13:26:41-04:00","recordedDate":"2001-08-31T13:26:41-04:00"},"request":{"method":"POST","url":"Condition"}},{"fullUrl":"urn:uuid:c16280ce-a07e-4e36-831d-1a3bfca8ffb4","resource":{"resourceType":"MedicationRequest","id":"c16280ce-a07e-4e36-831d-1a3bfca8ffb4","status":"active","intent":"order","medicationCodeableConcept":{"coding":[{"system":"http://www.nlm.nih.gov/research/umls/rxnorm","code":"849574","display":"Naproxen sodium 220 MG Oral Tablet"}],"text":"Naproxen sodium 220 MG Oral Tablet"},"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"},"authoredOn":"2001-08-31T13:26:41-04:00","requester":{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0","display":"Dr. Latricia419 Spencer878"},"reasonReference":[{"reference":"urn:uuid:a4e96085-e006-4f9a-8eb9-00bc41b0ab15"}]},"request":{"method":"POST","url":"MedicationRequest"}},{"fullUrl":"urn:uuid:7e8c884d-e6da-427b-83c4-3c0102ccb9af","resource":{"resourceType":"Claim","id":"7e8c884d-e6da-427b-83c4-3c0102ccb9af","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"pharmacy"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2001-08-31T13:26:41-04:00","end":"2001-08-31T13:41:41-04:00"},"created":"2001-08-31T13:41:41-04:00","provider":{"reference":"urn:uuid:2864a135-87f9-316b-a6b7-a272636846b6","display":"HIGH POINT REGIONAL HOSPITAL"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"prescription":{"reference":"urn:uuid:c16280ce-a07e-4e36-831d-1a3bfca8ffb4"},"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"185347001","display":"Encounter for problem"}],"text":"Encounter for problem"},"encounter":[{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"}]}],"total":{"value":11.37,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:82525203-566f-4022-86d1-1deabe8cc773","resource":{"resourceType":"CareTeam","id":"82525203-566f-4022-86d1-1deabe8cc773","status":"active","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"},"period":{"start":"2001-08-31T13:26:41-04:00"},"participant":[{"role":[{"coding":[{"system":"http://snomed.info/sct","code":"116153009","display":"Patient"}],"text":"Patient"}],"member":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"}},{"role":[{"coding":[{"system":"http://snomed.info/sct","code":"223366009","display":"Healthcare professional (occupation)"}],"text":"Healthcare professional (occupation)"}],"member":{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0","display":"Dr. Latricia419 Spencer878"}},{"role":[{"coding":[{"system":"http://snomed.info/sct","code":"224891009","display":"Healthcare services (qualifier value)"}],"text":"Healthcare services (qualifier value)"}],"member":{"reference":"urn:uuid:2864a135-87f9-316b-a6b7-a272636846b6","display":"HIGH POINT REGIONAL HOSPITAL"}}],"reasonCode":[{"coding":[{"system":"http://snomed.info/sct","code":"239873007","display":"Osteoarthritis of knee"}],"text":"Osteoarthritis of knee"}],"managingOrganization":[{"reference":"urn:uuid:2864a135-87f9-316b-a6b7-a272636846b6","display":"HIGH POINT REGIONAL HOSPITAL"}]},"request":{"method":"POST","url":"CareTeam"}},{"fullUrl":"urn:uuid:b4b0aded-bc9e-4b3c-aa69-19aad75a68ff","resource":{"resourceType":"CarePlan","id":"b4b0aded-bc9e-4b3c-aa69-19aad75a68ff","text":{"status":"generated"},"status":"active","intent":"order","category":[{"coding":[{"system":"http://snomed.info/sct","code":"408869004","display":"Musculoskeletal care"}],"text":"Musculoskeletal care"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"},"period":{"start":"2001-08-31T13:26:41-04:00"},"careTeam":[{"reference":"urn:uuid:82525203-566f-4022-86d1-1deabe8cc773"}],"addresses":[{"reference":"urn:uuid:a4e96085-e006-4f9a-8eb9-00bc41b0ab15"}],"activity":[{"detail":{"code":{"coding":[{"system":"http://snomed.info/sct","code":"386294003","display":"Joint mobility exercises"}],"text":"Joint mobility exercises"},"status":"in-progress","location":{"display":"HIGH POINT REGIONAL HOSPITAL"}}},{"detail":{"code":{"coding":[{"system":"http://snomed.info/sct","code":"266694003","display":"Heat therapy"}],"text":"Heat therapy"},"status":"in-progress","location":{"display":"HIGH POINT REGIONAL HOSPITAL"}}}]},"request":{"method":"POST","url":"CarePlan"}},{"fullUrl":"urn:uuid:21e3f21e-dbfa-44a4-8cd6-8837a490c11c","resource":{"resourceType":"Claim","id":"21e3f21e-dbfa-44a4-8cd6-8837a490c11c","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2001-08-31T13:26:41-04:00","end":"2001-08-31T13:41:41-04:00"},"created":"2001-08-31T13:41:41-04:00","provider":{"reference":"urn:uuid:2864a135-87f9-316b-a6b7-a272636846b6","display":"HIGH POINT REGIONAL HOSPITAL"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"diagnosis":[{"sequence":1,"diagnosisReference":{"reference":"urn:uuid:a4e96085-e006-4f9a-8eb9-00bc41b0ab15"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"185347001","display":"Encounter for problem"}],"text":"Encounter for problem"},"encounter":[{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"}]},{"sequence":2,"diagnosisSequence":[1],"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"239873007","display":"Osteoarthritis of knee"}],"text":"Osteoarthritis of knee"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:ff9fa160-712a-4647-bc0e-e0c68eccff25","resource":{"resourceType":"ExplanationOfBenefit","id":"ff9fa160-712a-4647-bc0e-e0c68eccff25","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0"},"performer":[{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"21e3f21e-dbfa-44a4-8cd6-8837a490c11c"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2001-08-31T13:41:41-04:00","end":"2002-08-31T13:41:41-04:00"},"created":"2001-08-31T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:21e3f21e-dbfa-44a4-8cd6-8837a490c11c"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:e47ccc52-d9b7-3589-b220-d0f9b38048d0"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"diagnosis":[{"sequence":1,"diagnosisReference":{"reference":"urn:uuid:a4e96085-e006-4f9a-8eb9-00bc41b0ab15"},"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-diagnosistype","code":"principal"}]}]}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"185347001","display":"Encounter for problem"}],"text":"Encounter for problem"},"servicedPeriod":{"start":"2001-08-31T13:26:41-04:00","end":"2001-08-31T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"21","display":"Inpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:6ed0d1c0-f189-42a3-b0bc-deb2a8ac72e8"}]},{"sequence":2,"diagnosisSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"239873007","display":"Osteoarthritis of knee"}],"text":"Osteoarthritis of knee"},"servicedPeriod":{"start":"2001-08-31T13:26:41-04:00","end":"2001-08-31T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"21","display":"Inpatient Hospital"}]}}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":0.0,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}},{"fullUrl":"urn:uuid:8ae7e67f-3628-4804-9251-dfcdb7343242","resource":{"resourceType":"Encounter","id":"8ae7e67f-3628-4804-9251-dfcdb7343242","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2014-09-22T13:26:41-04:00","end":"2014-09-22T13:41:41-04:00"},"individual":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc","display":"Dr. Emilio417 Swift555"}}],"period":{"start":"2014-09-22T13:26:41-04:00","end":"2014-09-22T13:41:41-04:00"},"serviceProvider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:3e5ae73f-7c04-48b3-b8db-177a15e7bdd1","resource":{"resourceType":"Immunization","id":"3e5ae73f-7c04-48b3-b8db-177a15e7bdd1","status":"completed","vaccineCode":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:8ae7e67f-3628-4804-9251-dfcdb7343242"},"occurrenceDateTime":"2014-09-22T13:26:41-04:00","primarySource":true},"request":{"method":"POST","url":"Immunization"}},{"fullUrl":"urn:uuid:f96a6121-997a-490a-812b-495073e34f6f","resource":{"resourceType":"Claim","id":"f96a6121-997a-490a-812b-495073e34f6f","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2014-09-22T13:26:41-04:00","end":"2014-09-22T13:41:41-04:00"},"created":"2014-09-22T13:41:41-04:00","provider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"supportingInfo":[{"sequence":1,"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claiminformationcategory","code":"info"}]},"valueReference":{"reference":"urn:uuid:3e5ae73f-7c04-48b3-b8db-177a15e7bdd1"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"encounter":[{"reference":"urn:uuid:8ae7e67f-3628-4804-9251-dfcdb7343242"}]},{"sequence":2,"informationSequence":[1],"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"net":{"value":106.67,"currency":"USD"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:c4e3f88d-066c-4745-8e70-54a7ab142109","resource":{"resourceType":"ExplanationOfBenefit","id":"c4e3f88d-066c-4745-8e70-54a7ab142109","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"performer":[{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"f96a6121-997a-490a-812b-495073e34f6f"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2014-09-22T13:41:41-04:00","end":"2015-09-22T13:41:41-04:00"},"created":"2014-09-22T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:f96a6121-997a-490a-812b-495073e34f6f"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"servicedPeriod":{"start":"2014-09-22T13:26:41-04:00","end":"2014-09-22T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:8ae7e67f-3628-4804-9251-dfcdb7343242"}]},{"sequence":2,"informationSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"servicedPeriod":{"start":"2014-09-22T13:26:41-04:00","end":"2014-09-22T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"net":{"value":106.67,"currency":"USD"},"adjudication":[{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_coinsrnc_amt","display":"Line Beneficiary Coinsurance Amount"}]},"amount":{"value":21.334000000000003,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prvdr_pmt_amt","display":"Line Provider Payment Amount"}]},"amount":{"value":85.33600000000001,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_sbmtd_chrg_amt","display":"Line Submitted Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_alowd_chrg_amt","display":"Line Allowed Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_bene_ptb_ddctbl_amt","display":"Line Beneficiary Part B Deductible Amount"}]},"amount":{"value":0,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prcsg_ind_cd","display":"Line Processing Indicator Code"}]}}]}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":85.33600000000001,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}},{"fullUrl":"urn:uuid:97822cbf-0f8f-4d93-8d34-377337228cbe","resource":{"resourceType":"Encounter","id":"97822cbf-0f8f-4d93-8d34-377337228cbe","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2015-09-28T13:26:41-04:00","end":"2015-09-28T13:41:41-04:00"},"individual":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc","display":"Dr. Emilio417 Swift555"}}],"period":{"start":"2015-09-28T13:26:41-04:00","end":"2015-09-28T13:41:41-04:00"},"serviceProvider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:13a4cef1-afd5-4aa3-b6e2-4c025b255496","resource":{"resourceType":"Immunization","id":"13a4cef1-afd5-4aa3-b6e2-4c025b255496","status":"completed","vaccineCode":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:97822cbf-0f8f-4d93-8d34-377337228cbe"},"occurrenceDateTime":"2015-09-28T13:26:41-04:00","primarySource":true},"request":{"method":"POST","url":"Immunization"}},{"fullUrl":"urn:uuid:fed79b01-62bd-45a7-b605-b178a24f5356","resource":{"resourceType":"Claim","id":"fed79b01-62bd-45a7-b605-b178a24f5356","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2015-09-28T13:26:41-04:00","end":"2015-09-28T13:41:41-04:00"},"created":"2015-09-28T13:41:41-04:00","provider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"supportingInfo":[{"sequence":1,"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claiminformationcategory","code":"info"}]},"valueReference":{"reference":"urn:uuid:13a4cef1-afd5-4aa3-b6e2-4c025b255496"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"encounter":[{"reference":"urn:uuid:97822cbf-0f8f-4d93-8d34-377337228cbe"}]},{"sequence":2,"informationSequence":[1],"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"net":{"value":106.67,"currency":"USD"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:1f5b9545-9d1d-45af-8f78-4f32da40148f","resource":{"resourceType":"ExplanationOfBenefit","id":"1f5b9545-9d1d-45af-8f78-4f32da40148f","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"performer":[{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"fed79b01-62bd-45a7-b605-b178a24f5356"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2015-09-28T13:41:41-04:00","end":"2016-09-28T13:41:41-04:00"},"created":"2015-09-28T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:fed79b01-62bd-45a7-b605-b178a24f5356"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"servicedPeriod":{"start":"2015-09-28T13:26:41-04:00","end":"2015-09-28T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:97822cbf-0f8f-4d93-8d34-377337228cbe"}]},{"sequence":2,"informationSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"servicedPeriod":{"start":"2015-09-28T13:26:41-04:00","end":"2015-09-28T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"net":{"value":106.67,"currency":"USD"},"adjudication":[{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_coinsrnc_amt","display":"Line Beneficiary Coinsurance Amount"}]},"amount":{"value":21.334000000000003,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prvdr_pmt_amt","display":"Line Provider Payment Amount"}]},"amount":{"value":85.33600000000001,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_sbmtd_chrg_amt","display":"Line Submitted Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_alowd_chrg_amt","display":"Line Allowed Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_bene_ptb_ddctbl_amt","display":"Line Beneficiary Part B Deductible Amount"}]},"amount":{"value":0,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prcsg_ind_cd","display":"Line Processing Indicator Code"}]}}]}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":85.33600000000001,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}},{"fullUrl":"urn:uuid:5b6d8fa3-ec50-4b38-84c6-f4b13741f6dd","resource":{"resourceType":"Encounter","id":"5b6d8fa3-ec50-4b38-84c6-f4b13741f6dd","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2016-10-03T13:26:41-04:00","end":"2016-10-03T13:41:41-04:00"},"individual":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc","display":"Dr. Emilio417 Swift555"}}],"period":{"start":"2016-10-03T13:26:41-04:00","end":"2016-10-03T13:41:41-04:00"},"serviceProvider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:bab6aa43-0a40-4c51-b6fc-594a0566915a","resource":{"resourceType":"Immunization","id":"bab6aa43-0a40-4c51-b6fc-594a0566915a","status":"completed","vaccineCode":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:5b6d8fa3-ec50-4b38-84c6-f4b13741f6dd"},"occurrenceDateTime":"2016-10-03T13:26:41-04:00","primarySource":true},"request":{"method":"POST","url":"Immunization"}},{"fullUrl":"urn:uuid:2fccd372-ccbd-4698-a254-749937625a6c","resource":{"resourceType":"Claim","id":"2fccd372-ccbd-4698-a254-749937625a6c","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2016-10-03T13:26:41-04:00","end":"2016-10-03T13:41:41-04:00"},"created":"2016-10-03T13:41:41-04:00","provider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"supportingInfo":[{"sequence":1,"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claiminformationcategory","code":"info"}]},"valueReference":{"reference":"urn:uuid:bab6aa43-0a40-4c51-b6fc-594a0566915a"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"encounter":[{"reference":"urn:uuid:5b6d8fa3-ec50-4b38-84c6-f4b13741f6dd"}]},{"sequence":2,"informationSequence":[1],"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"net":{"value":106.67,"currency":"USD"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:bf787c0c-2bbb-43b5-b364-abba6f0e87c0","resource":{"resourceType":"ExplanationOfBenefit","id":"bf787c0c-2bbb-43b5-b364-abba6f0e87c0","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"performer":[{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"2fccd372-ccbd-4698-a254-749937625a6c"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2016-10-03T13:41:41-04:00","end":"2017-10-03T13:41:41-04:00"},"created":"2016-10-03T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:2fccd372-ccbd-4698-a254-749937625a6c"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"servicedPeriod":{"start":"2016-10-03T13:26:41-04:00","end":"2016-10-03T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:5b6d8fa3-ec50-4b38-84c6-f4b13741f6dd"}]},{"sequence":2,"informationSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"servicedPeriod":{"start":"2016-10-03T13:26:41-04:00","end":"2016-10-03T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"net":{"value":106.67,"currency":"USD"},"adjudication":[{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_coinsrnc_amt","display":"Line Beneficiary Coinsurance Amount"}]},"amount":{"value":21.334000000000003,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prvdr_pmt_amt","display":"Line Provider Payment Amount"}]},"amount":{"value":85.33600000000001,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_sbmtd_chrg_amt","display":"Line Submitted Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_alowd_chrg_amt","display":"Line Allowed Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_bene_ptb_ddctbl_amt","display":"Line Beneficiary Part B Deductible Amount"}]},"amount":{"value":0,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prcsg_ind_cd","display":"Line Processing Indicator Code"}]}}]}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":85.33600000000001,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}},{"fullUrl":"urn:uuid:556063fc-bfa3-4264-a1a1-6a9ebc033e36","resource":{"resourceType":"Encounter","id":"556063fc-bfa3-4264-a1a1-6a9ebc033e36","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2017-10-09T13:26:41-04:00","end":"2017-10-09T13:41:41-04:00"},"individual":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc","display":"Dr. Emilio417 Swift555"}}],"period":{"start":"2017-10-09T13:26:41-04:00","end":"2017-10-09T13:41:41-04:00"},"serviceProvider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:e10ce12a-1368-4162-872f-8b6e3da50583","resource":{"resourceType":"Immunization","id":"e10ce12a-1368-4162-872f-8b6e3da50583","status":"completed","vaccineCode":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:556063fc-bfa3-4264-a1a1-6a9ebc033e36"},"occurrenceDateTime":"2017-10-09T13:26:41-04:00","primarySource":true},"request":{"method":"POST","url":"Immunization"}},{"fullUrl":"urn:uuid:cf098240-9b13-48c1-bb2a-f8b2777b9166","resource":{"resourceType":"Claim","id":"cf098240-9b13-48c1-bb2a-f8b2777b9166","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2017-10-09T13:26:41-04:00","end":"2017-10-09T13:41:41-04:00"},"created":"2017-10-09T13:41:41-04:00","provider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"supportingInfo":[{"sequence":1,"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claiminformationcategory","code":"info"}]},"valueReference":{"reference":"urn:uuid:e10ce12a-1368-4162-872f-8b6e3da50583"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"encounter":[{"reference":"urn:uuid:556063fc-bfa3-4264-a1a1-6a9ebc033e36"}]},{"sequence":2,"informationSequence":[1],"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"net":{"value":106.67,"currency":"USD"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:4a368c42-2533-4539-a03c-a8646dadd66b","resource":{"resourceType":"ExplanationOfBenefit","id":"4a368c42-2533-4539-a03c-a8646dadd66b","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"performer":[{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"cf098240-9b13-48c1-bb2a-f8b2777b9166"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2017-10-09T13:41:41-04:00","end":"2018-10-09T13:41:41-04:00"},"created":"2017-10-09T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:cf098240-9b13-48c1-bb2a-f8b2777b9166"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"servicedPeriod":{"start":"2017-10-09T13:26:41-04:00","end":"2017-10-09T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:556063fc-bfa3-4264-a1a1-6a9ebc033e36"}]},{"sequence":2,"informationSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"servicedPeriod":{"start":"2017-10-09T13:26:41-04:00","end":"2017-10-09T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"net":{"value":106.67,"currency":"USD"},"adjudication":[{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_coinsrnc_amt","display":"Line Beneficiary Coinsurance Amount"}]},"amount":{"value":21.334000000000003,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prvdr_pmt_amt","display":"Line Provider Payment Amount"}]},"amount":{"value":85.33600000000001,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_sbmtd_chrg_amt","display":"Line Submitted Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_alowd_chrg_amt","display":"Line Allowed Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_bene_ptb_ddctbl_amt","display":"Line Beneficiary Part B Deductible Amount"}]},"amount":{"value":0,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prcsg_ind_cd","display":"Line Processing Indicator Code"}]}}]}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":85.33600000000001,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}},{"fullUrl":"urn:uuid:937c2237-8e1c-4e2d-b0dd-24099e46ce85","resource":{"resourceType":"Encounter","id":"937c2237-8e1c-4e2d-b0dd-24099e46ce85","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2018-10-15T13:26:41-04:00","end":"2018-10-15T13:41:41-04:00"},"individual":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc","display":"Dr. Emilio417 Swift555"}}],"period":{"start":"2018-10-15T13:26:41-04:00","end":"2018-10-15T13:41:41-04:00"},"serviceProvider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:ba2bde85-a852-433c-80f4-5a26359421ba","resource":{"resourceType":"Immunization","id":"ba2bde85-a852-433c-80f4-5a26359421ba","status":"completed","vaccineCode":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:937c2237-8e1c-4e2d-b0dd-24099e46ce85"},"occurrenceDateTime":"2018-10-15T13:26:41-04:00","primarySource":true},"request":{"method":"POST","url":"Immunization"}},{"fullUrl":"urn:uuid:89d9fb43-6dca-4d31-99b1-59f047558f56","resource":{"resourceType":"Claim","id":"89d9fb43-6dca-4d31-99b1-59f047558f56","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2018-10-15T13:26:41-04:00","end":"2018-10-15T13:41:41-04:00"},"created":"2018-10-15T13:41:41-04:00","provider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"supportingInfo":[{"sequence":1,"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claiminformationcategory","code":"info"}]},"valueReference":{"reference":"urn:uuid:ba2bde85-a852-433c-80f4-5a26359421ba"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"encounter":[{"reference":"urn:uuid:937c2237-8e1c-4e2d-b0dd-24099e46ce85"}]},{"sequence":2,"informationSequence":[1],"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"net":{"value":106.67,"currency":"USD"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:b7322d0c-4728-45fd-8a32-6cfc45257d5c","resource":{"resourceType":"ExplanationOfBenefit","id":"b7322d0c-4728-45fd-8a32-6cfc45257d5c","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"performer":[{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"89d9fb43-6dca-4d31-99b1-59f047558f56"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2018-10-15T13:41:41-04:00","end":"2019-10-15T13:41:41-04:00"},"created":"2018-10-15T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:89d9fb43-6dca-4d31-99b1-59f047558f56"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"servicedPeriod":{"start":"2018-10-15T13:26:41-04:00","end":"2018-10-15T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:937c2237-8e1c-4e2d-b0dd-24099e46ce85"}]},{"sequence":2,"informationSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"servicedPeriod":{"start":"2018-10-15T13:26:41-04:00","end":"2018-10-15T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"net":{"value":106.67,"currency":"USD"},"adjudication":[{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_coinsrnc_amt","display":"Line Beneficiary Coinsurance Amount"}]},"amount":{"value":21.334000000000003,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prvdr_pmt_amt","display":"Line Provider Payment Amount"}]},"amount":{"value":85.33600000000001,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_sbmtd_chrg_amt","display":"Line Submitted Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_alowd_chrg_amt","display":"Line Allowed Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_bene_ptb_ddctbl_amt","display":"Line Beneficiary Part B Deductible Amount"}]},"amount":{"value":0,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prcsg_ind_cd","display":"Line Processing Indicator Code"}]}}]}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":85.33600000000001,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}},{"fullUrl":"urn:uuid:43f8b21d-8f4f-47f3-ba9b-f7eca33da8be","resource":{"resourceType":"Encounter","id":"43f8b21d-8f4f-47f3-ba9b-f7eca33da8be","status":"finished","class":{"system":"http://terminology.hl7.org/CodeSystem/v3-ActCode","code":"AMB"},"type":[{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"}],"subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Mrs. Theola421 Haag279"},"participant":[{"type":[{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/v3-ParticipationType","code":"PPRF","display":"primary performer"}],"text":"primary performer"}],"period":{"start":"2019-10-21T13:26:41-04:00","end":"2019-10-21T13:41:41-04:00"},"individual":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc","display":"Dr. Emilio417 Swift555"}}],"period":{"start":"2019-10-21T13:26:41-04:00","end":"2019-10-21T13:41:41-04:00"},"serviceProvider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"}},"request":{"method":"POST","url":"Encounter"}},{"fullUrl":"urn:uuid:9f49fa62-dff7-482c-9b2c-9979414ff03e","resource":{"resourceType":"Immunization","id":"9f49fa62-dff7-482c-9b2c-9979414ff03e","status":"completed","vaccineCode":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"encounter":{"reference":"urn:uuid:43f8b21d-8f4f-47f3-ba9b-f7eca33da8be"},"occurrenceDateTime":"2019-10-21T13:26:41-04:00","primarySource":true},"request":{"method":"POST","url":"Immunization"}},{"fullUrl":"urn:uuid:64e711ee-5e65-49ce-bc79-aa786198aaba","resource":{"resourceType":"Claim","id":"64e711ee-5e65-49ce-bc79-aa786198aaba","status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11","display":"Theola421 Haag279"},"billablePeriod":{"start":"2019-10-21T13:26:41-04:00","end":"2019-10-21T13:41:41-04:00"},"created":"2019-10-21T13:41:41-04:00","provider":{"reference":"urn:uuid:cf39a887-fde8-3674-898d-536b77c15bf0","display":"PCP148003"},"priority":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/processpriority","code":"normal"}]},"supportingInfo":[{"sequence":1,"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claiminformationcategory","code":"info"}]},"valueReference":{"reference":"urn:uuid:9f49fa62-dff7-482c-9b2c-9979414ff03e"}}],"insurance":[{"sequence":1,"focal":true,"coverage":{"display":"Cigna Health"}}],"item":[{"sequence":1,"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"encounter":[{"reference":"urn:uuid:43f8b21d-8f4f-47f3-ba9b-f7eca33da8be"}]},{"sequence":2,"informationSequence":[1],"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"net":{"value":106.67,"currency":"USD"}}],"total":{"value":98.05,"currency":"USD"}},"request":{"method":"POST","url":"Claim"}},{"fullUrl":"urn:uuid:d11a0103-da53-4f80-80a0-5aba2502e4a9","resource":{"resourceType":"ExplanationOfBenefit","id":"d11a0103-da53-4f80-80a0-5aba2502e4a9","contained":[{"resourceType":"ServiceRequest","id":"referral","status":"completed","intent":"order","subject":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"requester":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"performer":[{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"}]},{"resourceType":"Coverage","id":"coverage","status":"active","type":{"text":"Cigna Health"},"beneficiary":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"payor":[{"display":"Cigna Health"}]}],"identifier":[{"system":"https://bluebutton.cms.gov/resources/variables/clm_id","value":"64e711ee-5e65-49ce-bc79-aa786198aaba"},{"system":"https://bluebutton.cms.gov/resources/identifier/claim-group","value":"99999999999"}],"status":"active","type":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claim-type","code":"institutional"}]},"use":"claim","patient":{"reference":"urn:uuid:6aff2910-82fc-44d6-84a6-c29e4b756b11"},"billablePeriod":{"start":"2019-10-21T13:41:41-04:00","end":"2020-10-21T13:41:41-04:00"},"created":"2019-10-21T13:41:41-04:00","insurer":{"display":"Cigna Health"},"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"referral":{"reference":"#referral"},"claim":{"reference":"urn:uuid:64e711ee-5e65-49ce-bc79-aa786198aaba"},"outcome":"complete","careTeam":[{"sequence":1,"provider":{"reference":"urn:uuid:34ce0b83-7ceb-3496-8938-9dc3606a1bbc"},"role":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/claimcareteamrole","code":"primary","display":"Primary Care Practitioner"}]}}],"insurance":[{"focal":true,"coverage":{"reference":"#coverage","display":"Cigna Health"}}],"item":[{"sequence":1,"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://snomed.info/sct","code":"162673000","display":"General examination of patient (procedure)"}],"text":"General examination of patient (procedure)"},"servicedPeriod":{"start":"2019-10-21T13:26:41-04:00","end":"2019-10-21T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"encounter":[{"reference":"urn:uuid:43f8b21d-8f4f-47f3-ba9b-f7eca33da8be"}]},{"sequence":2,"informationSequence":[1],"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/variables/line_cms_type_srvc_cd","code":"1","display":"Medical care"}]},"productOrService":{"coding":[{"system":"http://hl7.org/fhir/sid/cvx","code":"140","display":"Influenza, seasonal, injectable, preservative free"}],"text":"Influenza, seasonal, injectable, preservative free"},"servicedPeriod":{"start":"2019-10-21T13:26:41-04:00","end":"2019-10-21T13:41:41-04:00"},"locationCodeableConcept":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/ex-serviceplace","code":"19","display":"Off Campus-Outpatient Hospital"}]},"net":{"value":106.67,"currency":"USD"},"adjudication":[{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_coinsrnc_amt","display":"Line Beneficiary Coinsurance Amount"}]},"amount":{"value":21.334000000000003,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prvdr_pmt_amt","display":"Line Provider Payment Amount"}]},"amount":{"value":85.33600000000001,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_sbmtd_chrg_amt","display":"Line Submitted Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_alowd_chrg_amt","display":"Line Allowed Charge Amount"}]},"amount":{"value":106.67,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_bene_ptb_ddctbl_amt","display":"Line Beneficiary Part B Deductible Amount"}]},"amount":{"value":0,"currency":"USD"}},{"category":{"coding":[{"system":"https://bluebutton.cms.gov/resources/codesystem/adjudication","code":"https://bluebutton.cms.gov/resources/variables/line_prcsg_ind_cd","display":"Line Processing Indicator Code"}]}}]}],"total":[{"category":{"coding":[{"system":"http://terminology.hl7.org/CodeSystem/adjudication","code":"submitted","display":"Submitted Amount"}],"text":"Submitted Amount"},"amount":{"value":98.05,"currency":"USD"}}],"payment":{"amount":{"value":85.33600000000001,"currency":"USD"}}},"request":{"method":"POST","url":"ExplanationOfBenefit"}}]}
   *      responses:
   *        "200":
   *          content:
   *            application/json:
   *              schema:
   *                type: object
   */

/**
    * @swagger
    *
    *   /api/BuildBundle/{_id}:
    *    post:
    *      summary: Build Bundle Around ID add filter build with certain FHIR Resources
    *      parameters:
    *          - in: path
    *            name: _id
    *            required: true
    *            example: 6aff2910-82fc-44d6-84a6-c29e4b756b11
    *      requestBody:
    *        required: false
    *        content:
    *           application/json:
    *             schema:
    *               type: object
    *             example: {"filter": "[\"Patient\", \"Condition\"]"}
    *      responses:
    *        "200":
    *          content:
    *            application/json:
    *              schema:
    *                type: object
    */

/**
     * @swagger
     *
     *   /api/delete:
     *    delete:
     *      summary: Delete all Nodes and Relationships from Neo4j
     *      responses:
     *        "200":
     *          content:
     *            application/json:
     *              schema:
     *                type: object
     */

/**
      * @swagger
      *
      *   /api/compareJSON:
      *    post:
      *      summary: Compare two JSON Objects to test CyFHIR
      *      requestBody:
      *        required: true
      *        content:
      *           application/json:
      *             schema:
      *               type: object
      *             example: {"jsons":[{"ex1":{"ex3":"value"},"ex2":"value"},{"ex2":"value","ex1":{"ex3":"value"}}]}
      *      responses:
      *        "200":
      *          content:
      *            application/json:
      *              schema:
      *                type: object
      */
